import asyncio
import unittest

from aioautobatch import autobatch


class AutoBatchTests(unittest.IsolatedAsyncioTestCase):
    async def test_batches_within_start_delay(self) -> None:
        calls: list[list[tuple[int, int]]] = []

        async def batch_fn(args_list: list[tuple[int, int]]) -> list[int]:
            calls.append(list(args_list))
            return [a + b for a, b in args_list]

        fn = autobatch(batch_fn, start_delay=0.05)

        futures = await asyncio.gather(*(fn(i, i + 1) for i in range(3)))
        results = await asyncio.gather(*futures)

        self.assertEqual(results, [1, 3, 5])
        self.assertEqual(len(calls), 1)
        self.assertEqual(calls[0], [(0, 1), (1, 2), (2, 3)])

    async def test_respects_batch_size(self) -> None:
        calls: list[list[tuple[int]]] = []

        async def batch_fn(args_list: list[tuple[int]]) -> list[int]:
            calls.append(list(args_list))
            return [args[0] for args in args_list]

        fn = autobatch(batch_fn, start_delay=0.05, batch_size=2, max_delay=0.01)

        futures = [await fn(i) for i in range(3)]
        results = await asyncio.gather(*futures)

        self.assertEqual(results, [0, 1, 2])
        self.assertEqual(len(calls), 2)
        self.assertEqual([len(batch) for batch in calls], [2, 1])

    async def test_limits_concurrent_batches(self) -> None:
        current = 0
        max_seen = 0

        async def batch_fn(args_list: list[tuple[int]]) -> list[int]:
            nonlocal current, max_seen
            current += 1
            max_seen = max(max_seen, current)
            await asyncio.sleep(0.05)
            current -= 1
            return [args[0] for args in args_list]

        fn = autobatch(batch_fn, batch_size=1, max_concurrent_batches=1)

        futures = await asyncio.gather(*(fn(i) for i in range(3)))
        results = await asyncio.gather(*futures)

        self.assertEqual(results, [0, 1, 2])
        self.assertEqual(max_seen, 1)

    async def test_exception_path_notifies_and_sets_future(self) -> None:
        caught: list[Exception] = []

        async def batch_fn(args_list: list[tuple[int]]) -> list[int]:
            raise ValueError("boom")

        def on_exception(exc: Exception) -> None:
            caught.append(exc)

        fn = autobatch(
            batch_fn, start_delay=0.02, batch_size=10, on_exception=on_exception
        )

        futures = await asyncio.gather(fn(1), fn(2))

        for fut in futures:
            with self.assertRaises(ValueError):
                await fut

        await asyncio.sleep(0)
        self.assertEqual(len(caught), 1)
        self.assertIsInstance(caught[0], ValueError)

    async def test_batch_fn_with_future_list_allows_early_resolution(self) -> None:
        completion_order: list[float] = []

        async def batch_fn(
            args_list: list[tuple[float]], futures: list[asyncio.Future[float]]
        ) -> None:
            for (delay,), fut in zip(args_list, futures):
                async def worker(d: float, f: asyncio.Future[float]) -> None:
                    await asyncio.sleep(d)
                    if f.cancelled() or f.done():
                        return
                    f.set_result(d)
                    completion_order.append(d)

                asyncio.create_task(worker(delay, fut))

            # Ensure worker tasks get a chance to start before we return
            await asyncio.sleep(0)

        fn = autobatch(batch_fn, start_delay=0.01, max_delay=0.01)

        slow = await fn(0.05)
        fast = await fn(0.01)

        fast_result = await asyncio.wait_for(fast, timeout=0.04)
        self.assertEqual(fast_result, 0.01)
        self.assertFalse(slow.done())

        slow_result = await asyncio.wait_for(slow, timeout=0.1)
        self.assertEqual(slow_result, 0.05)

        self.assertEqual(completion_order[0], 0.01)
        self.assertEqual(completion_order[-1], 0.05)



import typing
import dataclasses
import asyncio
import inspect

__all__ = ["autobatch"]

Ts = typing.TypeVarTuple("Ts")
R = typing.TypeVar("R")


class Fn(typing.Protocol[*Ts, R]):
    async def __call__(self, *args: typing.Unpack[Ts]) -> asyncio.Future[R]: ...


BatchFn: typing.TypeAlias = (
    typing.Callable[
        [list[tuple[typing.Unpack[Ts]]]],
        typing.Awaitable[list[R]],
    ]
    | typing.Callable[
        [list[tuple[typing.Unpack[Ts]]], list[asyncio.Future[R]]],
        typing.Awaitable[None],
    ]
)


@dataclasses.dataclass(frozen=True)
class _Job(typing.Generic[*Ts, R]):
    args: tuple[typing.Unpack[Ts]]
    future: asyncio.Future[R]
    created_at: float = dataclasses.field(
        default_factory=lambda: asyncio.get_running_loop().time()
    )


class _AutoBatcher(typing.Generic[*Ts, R]):
    def __init__(
        self,
        batch_fn: BatchFn[*Ts, R],
        start_delay: float,
        batch_size: int | None = None,
        max_delay: float = 1.0,
        max_concurrent_batches: int | None = None,
        on_exception: typing.Callable[[BaseException], None] | None = None,
    ) -> None:
        self._loop_task: None | asyncio.Task[None] = None
        self._batch_fn: BatchFn[*Ts, R] = batch_fn
        self._start_delay = start_delay
        self._batch_size = batch_size
        self._max_delay = max_delay
        self._batch_semaphore: asyncio.Semaphore | None = None
        if max_concurrent_batches is not None:
            self._batch_semaphore = asyncio.Semaphore(max_concurrent_batches)
        self._job_queue: asyncio.Queue[_Job[*Ts, R]] = asyncio.Queue()
        self._on_exception = on_exception
        self._accepts_future_list = _accepts_future_list(batch_fn)

    async def __call__(self, *args: typing.Unpack[Ts]) -> asyncio.Future[R]:
        job: _Job[*Ts, R] = _Job(
            args=args, future=asyncio.get_running_loop().create_future()
        )
        await self._job_queue.put(job)

        if self._loop_task is None:
            self._loop_task = asyncio.create_task(self._loop())

        return job.future

    async def _loop(self) -> None:
        try:
            await self._loop_main()
        finally:
            self._loop_task = None

    async def _loop_main(self) -> None:
        first = True
        while True:
            if self._job_queue.empty():
                # No jobs, wait for one
                return

            jobs: list[_Job[*Ts, R]] = []
            first_job = await self._job_queue.get()
            jobs.append(first_job)
            delay = self._start_delay if first else self._max_delay
            first = False
            await self._fetch_batch(jobs, delay)
            await self._process_batch(jobs)

    async def _process_batch(self, jobs: list[_Job[*Ts, R]]) -> None:
        if self._batch_semaphore is not None:
            await self._batch_semaphore.acquire()

        task = asyncio.create_task(self._execute_batch(jobs))
        task.add_done_callback(self._on_batch_done)

    async def _execute_batch(self, jobs: list[_Job[*Ts, R]]) -> None:
        args_list: list[tuple[typing.Unpack[Ts]]] = []
        futs: list[asyncio.Future[R]] = []
        for job in jobs:
            if job.future.cancelled():
                continue
            args_list.append(job.args)
            futs.append(job.future)
        try:
            if self._accepts_future_list:
                res_list = None
                await self._batch_fn(args_list, futs)
            else:
                res_list = await self._batch_fn(args_list)
        except Exception as e:
            for fut in futs:
                if not fut.cancelled() and not fut.done():
                    fut.set_exception(e)
            raise
        else:
            self._apply_results(futs, res_list)

    def _on_batch_done(self, task: asyncio.Task[None]) -> None:
        if self._batch_semaphore is not None:
            self._batch_semaphore.release()
        
        if task.cancelled() and self._on_exception is not None:
            # cancelled tasks do not have exceptions, so we create a generic one
            self._on_exception(asyncio.CancelledError())
            return
        
        exc = task.exception()
        if exc is not None and self._on_exception is not None:
            self._on_exception(exc)

    async def _fetch_batch(self, jobs: list[_Job[*Ts, R]], delay: float) -> None:
        begin = jobs[0].created_at

        while self._batch_size is None or len(jobs) < self._batch_size:
            now = asyncio.get_running_loop().time()
            elapsed = now - begin
            to_wait = delay - elapsed
            if to_wait <= 0:
                break

            try:
                job = await asyncio.wait_for(self._job_queue.get(), timeout=to_wait)
            except asyncio.TimeoutError:
                break

            jobs.append(job)

    @staticmethod
    def _apply_results(
        futs: list[asyncio.Future[R]], results: typing.Iterable[R] | None
    ) -> None:
        if results is None:
            return

        for fut, res in zip(futs, results):
            if fut.cancelled() or fut.done():
                continue
            fut.set_result(res)


def _accepts_future_list(batch_fn: typing.Callable[..., typing.Any]) -> bool:
    try:
        sig = inspect.signature(batch_fn)
    except (TypeError, ValueError):
        return False

    try:
        sig.bind_partial([], [])
    except TypeError:
        return False
    else:
        return True


def autobatch(
    batch_fn: BatchFn[*Ts, R],
    start_delay: float = 0,
    batch_size: int | None = None,
    max_delay: float = 1.0,
    max_concurrent_batches: int | None = None,
    on_exception: typing.Callable[[Exception], None] | None = None,
) -> Fn[*Ts, R]:
    return _AutoBatcher[*Ts, R](
        batch_fn,
        start_delay=start_delay,
        batch_size=batch_size,
        max_delay=max_delay,
        max_concurrent_batches=max_concurrent_batches,
        on_exception=on_exception,
    )
